﻿using EVCB_OCPP.Domain;
using EVCB_OCPP.Domain.Models.MainDb;
using EVCB_OCPP.Packet.Features;
using EVCB_OCPP.Packet.Messages;
using EVCB_OCPP.WSServer.Message;
using EVCB_OCPP.WSServer.Service;
using EVCB_OCPP.WSServer.Service.WsService;
using Microsoft.EntityFrameworkCore;
using Microsoft.Extensions.Configuration;
using Microsoft.Extensions.Logging;
using Newtonsoft.Json;
using Quartz;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace EVCB_OCPP.WSServer.Jobs;

[DisallowConcurrentExecution]
public class ServerMessageJob : IJob
{
    public ServerMessageJob(
        ProtalServer protalServer,
        ConfirmWaitingMessageSerevice confirmWaitingMessageSerevice,
        VendorIdReplaceService vendorIdReplaceService,
        IConfiguration configuration,
        IDbContextFactory<MainDBContext> maindbContextFactory,
        ILogger<ServerMessageJob> logger)
    {
        this.protalServer = protalServer;
        this.confirmWaitingMessageSerevice = confirmWaitingMessageSerevice;
        this.vendorIdReplaceService = vendorIdReplaceService;
        this.maindbContextFactory = maindbContextFactory;
        this.logger = logger;
    }

    private readonly ProtalServer protalServer;
    private readonly ConfirmWaitingMessageSerevice confirmWaitingMessageSerevice;
    private readonly VendorIdReplaceService vendorIdReplaceService;
    private readonly IDbContextFactory<MainDBContext> maindbContextFactory;
    private readonly ILogger<ServerMessageJob> logger;

    public async Task Execute(IJobExecutionContext context)
    {
        //logger.LogDebug("{0} Started", nameof(ServerMessageJob));
        try
        {
            await ExecuteTrigger();
        }
        catch (Exception ex)
        {
            logger.LogError("ServerMessageTrigger  Ex:{0}", ex.ToString());
        }
    }

    private Task ExecuteTrigger()
    {
        var clientDic = protalServer.GetClientDic();

        confirmWaitingMessageSerevice.RemoveExpiredConfirmMessage();
        ResendServerMessage(clientDic);
        return SendNewServerMessage(clientDic);
    }

    private async Task SendNewServerMessage(Dictionary<string, WsClientData> clientDic)
    {
        BasicMessageHandler msgAnalyser = new BasicMessageHandler();
        DateTime dateTimeNow = DateTime.UtcNow;
        DateTime startDt = dateTimeNow.AddSeconds(-30);
        DateTime dt = new DateTime(1991, 1, 1);
        DateTime currentTime = dateTimeNow;
        List<ServerMessage> commandList;

        using (var db = await maindbContextFactory.CreateDbContextAsync())
        {
            commandList = await db.ServerMessage.Where(c => c.ReceivedOn == dt && c.UpdatedOn == dt && c.CreatedOn >= startDt && c.CreatedOn <= currentTime).AsNoTracking().ToListAsync();
        }
        //處理主機傳送的有指令
        var cmdMachineList = commandList.Select(c => c.ChargeBoxId).Distinct().ToList();
        if (commandList.Count > 0)
        {
            // Console.WriteLine(string.Format("Now:{0} commandList Count:{1} ", DateTime.UtcNow.ToString("yyyy/MM/dd HH:mm:ss"), commandList.Count));
        }
        foreach (var charger_SN in cmdMachineList)
        {
            WsClientData session;
            string uuid = string.Empty;

            if (!clientDic.TryGetValue(charger_SN, out session))
            {
                continue;
            }

            //logger.LogDebug(string.Format("charger_SN:{0} startDt:{1} CreatedOn:{2}", charger_SN, startDt.ToString("yyyy/MM/dd HH:mm:ss"), DateTime.UtcNow.ToString("yyyy/MM/dd HH:mm:ss")));

            //if (session.IsCheckIn && !session.ISOCPP20)
            if (session.ISOCPP20)
            {
                continue;
            }

            string rawRequest = string.Empty;

            var cmdList = commandList.Where(c => c.ChargeBoxId == charger_SN).ToList();

            var profiles = protalServer.Profiles;

            foreach (var item in cmdList)
            {
                IRequest request = null;
                Actions action = Actions.None;
                Enum.TryParse(item.OutAction, out action);
                Type _RequestType = null;

                for (int i = 0; i < profiles.Count; i++)
                {
                    var feature = profiles[i].GetFeaturebyAction(item.OutAction);

                    if (feature != null)
                    {
                        _RequestType = feature.GetRequestType();
                        break;
                    }
                }

                if (_RequestType != null && item.CreatedBy != "Destroyer")
                {
                    request = JsonConvert.DeserializeObject(item.OutRequest, _RequestType) as IRequest;
                    uuid = session.queue.store(request);
                    rawRequest = BasicMessageHandler.GenerateRequest(uuid, item.OutAction, request);
                    protalServer.SendMsg(session, rawRequest, string.Format("{0} {1}", action, "Request"), "");
                }


                if (item.CreatedBy == "Destroyer")
                {

                    if (_RequestType != null)
                    {
                        request = Activator.CreateInstance(_RequestType) as IRequest;
                        uuid = session.queue.store(request);
                        rawRequest = BasicMessageHandler.GenerateDestroyRequest(uuid, item.OutAction, item.OutRequest);
                        protalServer.SendMsg(session, rawRequest, string.Format("{0} {1}", action, "Request"), "");

                    }
                    else
                    {

                        rawRequest = BasicMessageHandler.GenerateDestroyRequest(Guid.NewGuid().ToString(), item.OutAction, item.OutRequest);
                        protalServer.SendMsg(session, rawRequest, string.Format("{0} {1}", action, "Request"), "");

                    }
                }

                await confirmWaitingMessageSerevice.Add(charger_SN, item.Id, item.SerialNo, item.OutAction, uuid, item.CreatedBy, rawRequest);
                //protalServer.AddConfirmMessage(charger_SN, item.Id, item.SerialNo, item.OutAction, uuid, item.CreatedBy, rawRequest);

                await Task.Delay(100);

            }
        }
    }

    private void ResendServerMessage(Dictionary<string, WsClientData> clientDic)
    {
        DateTime dateTimeNow = DateTime.UtcNow;
        List<NeedConfirmMessage> resendList = confirmWaitingMessageSerevice.GetPendingMessages();
        foreach (var resendItem in resendList)
        {
            WsClientData session;
            if (clientDic.TryGetValue(resendItem.ChargePointSerialNumber, out session))
            {
                if (dateTimeNow.Subtract(resendItem.SentOn).TotalSeconds > 1)
                {
                    confirmWaitingMessageSerevice.SignalMessageSended(resendItem);
                    protalServer.SendMsg(session, resendItem.SentMessage, string.Format("{0} {1}", resendItem.SentAction, "Request"), "");
                }

            }

        }
    }
}
